/*
 * Contains pre-computed data used with a Physical BSDF.
 */
struct PhysicalBSDF {
    mediump vec3 diffuse;
    mediump vec3 specular;
    mediump float perceptualRoughness;
    /** @todo: Pre-compute the data re-used in distribution/geometry,
     * such as the roughness^2 */
};

/**
 * Create the BSDF data used with physical lighting equations
 *
 * @param albedo Surface albedo. Shouldn't be premultiplied
 *     by (1 - reflectivty)
 * @param metallic Surface metallic, in [0; 1]
 * @param roughness Surface **perceptual** roughness, in [0; 1]
 * @returns A structure containing BSDF pre-computed data
 */
PhysicalBSDF createPhysicalBSDF(vec3 albedo, float metallic, float roughness) {
    PhysicalBSDF data;
    data.diffuse = albedo*(1.0 - metallic)*RECIPROCAL_PI;
    data.specular = mix(vec3(0.04), albedo.rgb, metallic);
    data.perceptualRoughness = roughness;
    return data;
}

vec3 fresnelSchlick(float cosTheta, vec3 F0) {
    return F0 + (vec3(1.0) - F0)*pow(1.0 - cosTheta, 5.0);
}

float distributionGGX(vec3 N, vec3 H, float roughness) {
    float a = roughness*roughness;
    float a2 = a*a;
    float NdotH = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;

    float num = a2;
    float denom = (NdotH2*(a2 - 1.0) + 1.0);
    denom = PI*denom*denom;

    return num/denom;
}

float geometrySchlickGGX(float NdotV, float roughness) {
    float a = roughness*roughness;
    float r = (a + 1.0);
    float k = (r*r) / 8.0;

    float num = NdotV;
    float denom = NdotV*(1.0 - k) + k;

    return num/denom;
}

float geometrySmith(vec3 N, vec3 V, vec3 L, float roughness) {
    float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx2 = geometrySchlickGGX(NdotV, roughness);
    float ggx1 = geometrySchlickGGX(NdotL, roughness);

    return ggx1*ggx2;
}

vec3 physicalBrdf(PhysicalBSDF bsdf, mediump vec3 L, mediump vec3 normal, mediump vec3 view) {
    vec3 halfVec = normalize(view + L);

    float NDF = distributionGGX(normal, halfVec, bsdf.perceptualRoughness);
    float G = geometrySmith(normal, view, L, bsdf.perceptualRoughness);
    vec3 f = fresnelSchlick(max(dot(halfVec, view), 0.0), bsdf.specular);

    vec3 kD = (vec3(1.0) - f);

    vec3 numerator = NDF*G*f;
    float NdotL = max(dot(normal, L), 0.0);
    float denominator = 4.0*max(dot(normal, view), 0.0)*NdotL;
    vec3 specular = numerator/max(denominator, 0.001);

    return (kD*bsdf.diffuse + specular)*NdotL;
}

/**
 * Evaluate the diffuse and specular contributions of ponctual lights
 *
 * @param bsdf BSDF data, created via @ref createPhysicalBSDF()
 * @param view The view vector, in **world space**
 * @param normal The normal vector, in **world space**
 * @returns The diffuse and specular contribution
 */
vec3 evaluateDirectLights(PhysicalBSDF bsdf, mediump vec3 view, mediump vec3 normal) {
    vec3 col = vec3(0.0);

    #if NUM_LIGHTS > 0

    lowp uint i = 0u;
    for(; i < pointLightCount; ++i) {
        mediump vec4 lightData = lightColors[i];
        /* dot product of mediump vec3 can be NaN for distances > 128 */
        highp vec3 lightPos = lightPositionsWorld[i];
        highp vec3 lightDirAccurate = lightPos - fragPositionWorld;
        mediump float distSq = dot(lightDirAccurate, lightDirAccurate);
        mediump float attenuation = distanceAttenuation(distSq, lightData.a);
        if(attenuation < 0.001)
            continue;

        mediump vec3 lightDir = lightDirAccurate;
        lightDir *= inversesqrt(distSq);

        mediump vec3 value = physicalBrdf(bsdf, lightDir, normal, view);

        float shadow = 1.0;
        #if NUM_SHADOWS > 0
        /* Shadows */
        bool shadowsEnabled = bool(lightParameters[i].z);
        if(shadowsEnabled) {
            int shadowIndex = int(lightParameters[i].w) + int(dot(lightDir, lightDirectionsWorld[i]) < 0.0);
            shadow = sampleShadowParaboloid(shadowIndex);
        }
        #endif

        col += attenuation*value*lightData.rgb*shadow;
    }

    lowp uint endSpotLights = pointLightCount + spotLightCount;
    for(; i < endSpotLights; ++i) {
        mediump vec4 lightData = lightColors[i];
        /* dot product of mediump vec3 can be NaN for distances > 128 */
        highp vec3 lightPos = lightPositionsWorld[i];
        highp vec3 lightDirAccurate = lightPos - fragPositionWorld;
        mediump float distSq = dot(lightDirAccurate, lightDirAccurate);
        mediump float attenuation = distanceAttenuation(distSq, lightData.a);

        if(attenuation < 0.001)
            continue;

        mediump vec3 lightDir = lightDirAccurate;
        lightDir *= inversesqrt(distSq);

        highp vec3 spotDir = lightDirectionsWorld[i];
        attenuation *= spotAttenuation(lightDir, spotDir, lightParameters[i].x, lightParameters[i].y);

        if(attenuation < 0.001)
            continue;

        mediump vec3 value = physicalBrdf(bsdf, lightDir, normal, view);

        float shadow = 1.0;
        #if NUM_SHADOWS > 0
        /* Shadows */
        bool shadowsEnabled = bool(lightParameters[i].z);
        if(shadowsEnabled) {
            int shadowIndex = int(lightParameters[i].w);
            shadow = sampleShadowPerspective(shadowIndex, normal, lightDir);
        }
        #endif

        col += attenuation*value*lightData.rgb*shadow;
    }

    lowp uint endSunLights = pointLightCount + spotLightCount + sunLightCount;
    for(; i < endSunLights; ++i) {
        mediump vec4 lightData = lightColors[i];
        mediump vec3 lightDir = lightDirectionsWorld[i];

        mediump vec3 value = physicalBrdf(bsdf, lightDir, normal, view);

        float shadow = 1.0;
        #if NUM_SHADOWS > 0
        /* Shadows */
        bool shadowsEnabled = bool(lightParameters[i].z);
        if(shadowsEnabled) {
            int shadowIndex = int(lightParameters[i].w);
            float depth = -fragPositionView.z;
            int cascade = selectCascade(shadowIndex, depth);
            if(cascade != -1)
                shadow = sampleShadowOrtho(shadowIndex + cascade, normal, lightDir);
        }
        #endif

        col += value*lightData.a*lightData.rgb*shadow;
    }

    #endif

    return col;
}
