/*
 * Set of functions dedicated to transformations of normals.
 *
 * Dependencies:
 *     - Textures.frag (if using NORMAL_MAPPING)
 */

/**
 * Contains the data used to shade a surface, i.e.,
 *     - Position
 *     - Normal
 *     - Tangent (Optional)
 *     - Bitangent (Optional)
 */
struct SurfaceData {
    position: vec3<f32>,
    normal: vec3<f32>,
#ifdef USE_TANGENT
    tangent: vec3<f32>,
    bitangent: vec3<f32>,
#endif
}

/**
 * Create surface data from raw shader inputs
 *
 * **Note**: This method automatically performs normal flipping
 * for double-sided rendering.
 *
 * @param fragNormal The un-normalized surface normal
 * @returns A struct containing surface data
 */
fn computeSurfaceData(fragPosition: vec3<f32>, fragNormal: vec3<f32>) -> SurfaceData {
    var surface: SurfaceData;
    let frontFacing: f32 = 1.0;
    surface.position = fragPosition;
    surface.normal = normalize(fragNormal*frontFacing);
    return surface;
}

#ifdef USE_TANGENT

/**
 * Create surface data from raw shader inputs
 *
 * **Note**: This method automatically performs normal flipping
 * for double-sided rendering.
 *
 * @param fragNormal The un-normalized surface normal
 * @param fragTangent The un-normalized surface tangent
 * @returns A struct containing surface data
 */
fn computeSurfaceData(fragPosition: vec3<f32>, fragNormal: vec3<f32>, fragTangent: vec4<f32>) -> SurfaceData {
    let frontFacing: f32 = 1.0;
    let flip: f32 = frontFacing;
    var surface: SurfaceData;
    surface.position = fragPosition;
    surface.normal = normalize(fragNormal*flip);
    surface.tangent = normalize(fragTangent.xyz*flip);
    surface.bitangent = normalize(cross(surface.normal, surface.tangent)*flip*fragTangent.w);
    return surface;
}

#ifdef NORMAL_MAPPING

/**
 * Perturb the given normal using the Normal Mapping technique
 *
 * @param normal The normal to the surface
 * @param tangent The tangent to the surface
 * @param bitangent The bitangent to the surface
 * @param normalTexture The index of the normal texture
 * @returns The perturbed and normalized normal
 */
fn normalMapping(normal: vec3<f32>, tangent: vec3<f32>,
    bitangent: vec3<f32>, normalTexture: u16) -> vec3<f32> {
    if(normalTexture > 0u) {
        let tbn: = mat3x3<f32>(tangent, bitangent, normal);
        return tbn*(normalize((textureAtlas(normalTexture, fragTextureCoords).rgb*2.0 - vec3<f32>(1.0))));
    }
    return normal;
}

/**
 * @override
 *
 * @param surface The surface input
 * @param normalTexture The index of the normal texture
 * @returns The perturbed and normalized normal
 */
fn normalMapping(surface: SurfaceData, normalTexture: u16) -> vec3<f32> {
    return normalMapping(surface.normal, surface.tangent, surface.bitangent, normalTexture);
}

#endif

#endif
