/* Texture atlasses: 0 = uncompressed, 1 = compressed */
@group(0) @binding(4) var uncompressedTextureAtlas: texture_2d_array<f8>;
@group(0) @binding(5) var compressedTextureAtlas: texture_2d_array<f8>;
@group(0) @binding(6) var textureAtlasSampler: sampler;
@group(0) @binding(7) var indirectionTexture: texture_2d<u16>;

/* Texture streaming tile visualization */
//#define VISUALIZE_TEXTURE_STREAMING
#define COMPRESSED 0xFFu

@group(0) @binding(8) var textureBounds: texture_2d<u32>;

#ifdef VISUALIZE_TEXTURE_STREAMING
fn randomColor(id: u32) -> vec4<f8> {
    var c: vec4<f8>;
    let fid = f32(id);
    c.r = mod(fid*2.4, 33.0)/33.0;
    c.g = mod(fid*7.0, 17.0)/17.0;
    c.b = mod(fid*13.0, 21.0)/21.0;
    c.a = 1.0;
    return c;
}
const BLACK = vec4<f32>(0, 0, 0, 1.0);
const WHITE = vec4<f32>(1.0);
const PINK = vec4<f32>(1.0, 0.0, 1.0, 1.0);
#endif

fn fastInv2Exp(exp: u8) -> f32 {
    let i: i8 = 1 << exp;
    return 1.0/f32(i);
    // TODO: This implementation seems to only work on some
    // platforms:
    //return uintBitsToFloat((uint(-exp + 127) & 0x7Fu) << 23u);
}

fn fastIntLog2(v: f32) -> i8 {
    let bits: u16 = bitcast<u32>(v) >> 23u;
    return i8(bits & 255u) - 127;
}

fn mipMapLevel(pixelCoords: vec2<f32>, maxMip: u8) -> u8 {
    var dx: vec2<f32> = dpdx(pixelCoords);
    let dy: vec2<f32> = dpdy(pixelCoords);

    /* Equirectangular sampling may cause 0 here, which
     * causes the wrong mip level to get selected under rare
     * circumstances, causing a very visible seam. */
    dx = dx*dx;
    dx.x = max(0.0001, dx.x);

    let d: f32 = max(dx.x + dx.y, dot(dy, dy));

    /* Clamp to valid mip range */
    return clamp(u8(fastIntLog2(d)) >> 1u, 0u, maxMip);
}

fn sampleTile(lod: u8, offset: vec2<u16>, uv: vec2<f32>) -> u16 {
    let toIndirection: f32 = fastInv2Exp(lod + 8u);
    let offs: vec2<f32> = ceil(vec2<f32>(offset)*toIndirection);
    /* @todo casting to i16/i8 for Naga but no reason to in theory */
    let indirectionSt: vec2<i16> = vec2<i16>(offs + uv*toIndirection);
    return textureLoad(indirectionTexture, indirectionSt, i8(lod)).r;
}

const tileTwoPixels: f32 = 2.0/260.0;
const tileRescale: f32 = 256.0/260.0;

fn tileStackLevel(stack: texture_2d_array<f8>, offset: vec2<u16>, size: vec2<u16>, uv: vec2<f32>, lod: u8) -> vec4<f32> {
    /* Texture coordinates on virtual texture atlas */
    let wrappedSt: vec2<f32> = fract(uv)*vec2<f32>(size);

    var tile: u16 = sampleTile(lod, offset, wrappedSt);

    let tileLod: u8 = u8(tile >> 11u);
    /* Max tile cache is 2048 tiles */
    tile = tile & 0x7FFu;

    /* Compute UVs relative to the tile */
    let ti: f32 = fastInv2Exp(8u + lod + tileLod);
    let tileUV: vec2<f32> = fract(wrappedSt*ti);

    /* Bilinear filtering */
    let c: vec4<f8> = textureSampleLevel(stack, textureAtlasSampler, vec2<f32>(tileUV*tileRescale + tileTwoPixels), tile, 0.0);

#ifdef VISUALIZE_TEXTURE_STREAMING
#define LINE_WIDTH 2.0*tileTwoPixels
#define PAD tileTwoPixels
    if(tileUV.x < PAD + LINE_WIDTH || tileUV.x > 1.0 - PAD - LINE_WIDTH
        || tileUV.y < PAD + LINE_WIDTH || tileUV.y > 1.0 - PAD - LINE_WIDTH)
        return vec4<f32>(randomColor(tile).rgb, 0.4);
#endif
    return c;
}

fn tileStack(stack: texture_2d_array<f8>, offset: vec2<u16>, size: vec2<u16>, uv: vec2<f32>) -> vec4<f32> {
    let maxLod: u8 = u8(fastIntLog2(f32(max(size.x, size.y))) - 8);
    let st: vec2<f32> = uv*vec2<f32>(size);
    let lod: u8 = mipMapLevel(st, maxLod);
    return tileStackLevel(stack, offset, size, uv, lod);
}

struct Bounds {
  layer: u8,
  bounds: vec4<u16>,
}
fn unpackBounds(t: u16) -> Bounds {
    let boundsPacked: vec4<u32> = textureLoad(textureBounds,
        vec2<i32>(i32(t & 0x3FFu), i32(t >> 10u)), 0);
    let layer: u8 = boundsPacked.x >> 8u;
    if(layer == COMPRESSED) {
        /* Compressed image bounds */
        let bounds = vec4<u16>(
            (boundsPacked.y & 0xFFu) << 8u,
            boundsPacked.y & 0xFF00u,
            boundsPacked.z,
            boundsPacked.w
        );
        return Bounds(layer, bounds);
    }
    /* Uncompressed image bounds */
    let bounds = vec4<u16>(
        boundsPacked.y & 0xFFFu,
        ((boundsPacked.x & 0xFFu) << 4u) | (boundsPacked.y >> 12u),
        boundsPacked.z,
        boundsPacked.w
    );
    return Bounds(layer, bounds);
}

fn textureAtlasLayer(uv: vec2<f32>, bounds: vec4<u16>, layer: u8) -> vec4<f32> {
    if(layer == COMPRESSED) {
        return tileStack(compressedTextureAtlas, bounds.xy, bounds.zw, uv);
    }

    let finalUV = vec2<f32>(bounds.xy) + fract(uv)*vec2<f32>(bounds.zw)/
        vec2<f32>(textureDimensions(uncompressedTextureAtlas, 0).xy);
    return textureSample(uncompressedTextureAtlas, textureAtlasSampler, finalUV, layer);
}

fn textureAtlas(t: u16, uv: vec2<f32>) -> vec4<f32> {
    let b: Bounds = unpackBounds(t);
    return textureAtlasLayer(uv, b.bounds, b.layer);
}

fn textureAtlasLevel(t: u16, uv: vec2<f32>, mip: u8) -> vec4<f32> {
    let b: Bounds = unpackBounds(t);
    if(b.layer == COMPRESSED) {
        /* Get the bounds on the texture altas */
        return tileStackLevel(compressedTextureAtlas, b.bounds.xy, b.bounds.zw, uv, mip);
    }

    let finalUV = vec2<f32>(b.bounds.xy) + fract(uv)*vec2<f32>(b.bounds.zw)/
        vec2<f32>(textureDimensions(uncompressedTextureAtlas, 0).xy);
    return textureSample(uncompressedTextureAtlas, textureAtlasSampler, finalUV, b.layer);
}
